
/****************************************************************************
 National Instruments Corporation.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 ****************************************************************************/

#ifndef _PPB_SHIFT_H
#define _PPB_SHIFT_H

static __s32 max_allowed_ppb = SI514_DEFAULT_ADJ_PPB;

/****************************************************************************
 * Functions Declaration *
 ****************************************************************************/

static __int128_t div128_with_rest(__int128_t dividend, __int128_t divisor, __int128_t* remainder)
{
	__int128_t local_remainder = 0;
	__int128_t quotient = 0;
	__uint128_t mask = (__uint128_t)1 << 127;
	__s32 times = 127;

	// We first treat the special cases
	if(divisor == 0 || dividend == 0)
	{
		*remainder = 0;
		return 0;
	}
	else if(divisor == 1)
	{
		*remainder = 0;
		return dividend;
	}
	else if(dividend == divisor)
	{
		*remainder = 0;
		return 1;
	}
	else if(divisor > dividend)
	{
		*remainder = dividend;
		return 0;
	}

	//Get to the most significant bit
	while((mask & dividend) == 0)
	{
		mask = mask >> 1;
		times--;
	}

	//This is the long division algorithm
	for(; times >= 0; times--)
	{
		local_remainder = (local_remainder << 1) + ((mask & dividend) != 0 ? (1) : (0));
		if( local_remainder >= divisor)
		{
			local_remainder -= divisor;
			quotient += mask;
		}
		mask = mask >> 1;
	}
	*remainder = local_remainder;
	return quotient;
}

static __int128_t div128(__int128_t dividend, __int128_t divisor)
{
	__int128_t remainder = 0;
	return div128_with_rest(dividend, divisor, &remainder);
}

static struct tSi514_muldiv compute_ppb_shift_math(struct tSi514_muldiv dev_ref, __s32 dsp_int, __s32 dsp_frac)
{
	__int128_t new_int;
	__int128_t new_frac;
	__int128_t common_div;
	__int128_t sum;
	struct tSi514_muldiv new_muldiv;

	/*------ Shift Frequency -------*/
	/* starting from new_m = reference_m * div_freq_d; and replacing the values, we arrive to this formula
	new_m = dev_ref.m_int + ((dev_ref.m_frac  * si514_Ppb_divisor * fraction_scale) + (dev_ref.m_int * dsp_int * fraction_scale * si514_29bit_val) +
	(dev_ref.m_int * dsp_frac * si514_29bit_val) + (dev_ref.m_frac  * dsp_int * fraction_scale) + (dev_ref.m_frac  * dsp_frac)) / (si514_Ppb_divisor * fraction_scale * si514_29bit_val)
	*/
	common_div = ((__int128_t)si514_Ppb_divisor * fraction_scale * si514_29bit_val);
	sum = (((__int128_t)dev_ref.m_frac * si514_Ppb_divisor * fraction_scale) + \
	((__int128_t)dev_ref.m_int * dsp_int * fraction_scale * si514_29bit_val) + \
	((__int128_t)dev_ref.m_int * dsp_frac * si514_29bit_val) + \
	((__int128_t)dev_ref.m_frac * dsp_int * fraction_scale) + \
	((__int128_t)dev_ref.m_frac * dsp_frac));

	/* Taking the integer part only */
	new_int = dev_ref.m_int + div128_with_rest(sum, common_div,&new_frac);

	/* We simplify the value that was used to multiply later in the original code, the si514_29bit_val */
	common_div = (__int128_t)si514_Ppb_divisor * fraction_scale;

	/* Taking the fraction part only (29-bit resolution) */
	new_frac = div128(new_frac, common_div);

	new_muldiv = dev_ref;
	new_muldiv.m_int = (__u16)new_int;
	new_muldiv.m_frac = (__u32)new_frac;

	return new_muldiv;
}

/* ------------------- Device Specific functions --------------------- */

static int si514_compute_ppb_shift(struct i2c_client *client,
	struct tSi514_muldiv *new_freq_settings,
	__s32 desired_shift_in_ppb_int,
	__s32 desired_shift_in_ppb_fract)
{
	struct si514_data *dev_data;
	struct tSi514_muldiv new_muldiv;
	__u32 si514_m_max;
	__u64 si514_m_max_frac;
	__u32 si514_m_min;
	__u64 si514_m_min_frac;

	dev_data = i2c_get_clientdata(client);

	/* Intel x86_64 platform specific.
	Used because we have 29 fraction that need total of 128-bit to
	effectively divide and multiply.
	*/

	// While comparing, we only check the integer part and if they are equals, is enough
	// to check that the fractional part is not zero to see if it's the bigger number.
	if (desired_shift_in_ppb_int > max_allowed_ppb || (desired_shift_in_ppb_int == max_allowed_ppb && desired_shift_in_ppb_fract != 0) ||
		desired_shift_in_ppb_int < -max_allowed_ppb || (desired_shift_in_ppb_int == -max_allowed_ppb && desired_shift_in_ppb_fract != 0)) {
		dev_warn(&client->dev,
			"\n Desired PPB INT = %ld, is out of range",
			(long) desired_shift_in_ppb_int);
		dev_warn(&client->dev,
			"\n Max PPB INT = %d, is out of range",
			max_allowed_ppb);
		return -ERANGE;
	}

	/* Compute the frequency shift from the initial state frequency */
	new_muldiv = compute_ppb_shift_math(
		*dev_data->reference_hw_settings,
		desired_shift_in_ppb_int,
		desired_shift_in_ppb_fract);

	/* Range Validation, check before writing to the HW */
	si514_m_max = 78; /* From data sheet page #12 */
	si514_m_max_frac = 93339657;
	si514_m_min = 65; /* From data sheet page #12 */
	si514_m_min_frac = 21824022;
	// Comparing the integer part first and if they are equal, compare the fractional part.
	if (new_muldiv.m_int > si514_m_max || (new_muldiv.m_int == si514_m_max && new_muldiv.m_frac > si514_m_max_frac) 
	|| new_muldiv.m_int < si514_m_min || (new_muldiv.m_int == si514_m_min && new_muldiv.m_frac < si514_m_min_frac)) {
		dev_warn(&client->dev,
			"\n Failed : Out of range, computed M = %d,%u\n",
			new_muldiv.m_int, new_muldiv.m_frac);
		return -EIO;
	}

	*new_freq_settings = new_muldiv;

	return 0;
}

/*///////////////////////////////////////////////////////////////////////////*/

static void compute_current_ppb_shift(struct si514_data *dev_data,
	__s32 *ppb_int_part, __s32 *ppb_fract_part)
{
	__int128_t curr_int = dev_data->current_hw_settings.m_int;
	__int128_t curr_frac = dev_data->current_hw_settings.m_frac;
	__int128_t curr_dec = (__int128_t)dev_data->current_hw_settings.hs_div * (1ll << dev_data->current_hw_settings.ls_div_bits);
	__int128_t ref_int = dev_data->reference_hw_settings->m_int;
	__int128_t ref_frac = dev_data->reference_hw_settings->m_frac;
	__int128_t ref_dec = (__int128_t)dev_data->reference_hw_settings->hs_div * (1ll << dev_data->reference_hw_settings->ls_div_bits);
	__int128_t result_int = 0;
	__int128_t result_frac = 0;
	__int128_t common_div = (ref_int * si514_29bit_val + ref_frac) * curr_dec;
	__int128_t sum  = ((curr_int * si514_29bit_val + curr_frac) * ref_dec - common_div) * si514_Ppb_divisor;

	/* Starting from ppb_shifted = (double) ((double) freq_shift / ref_freq_hz) * si514_Ppb_divisor; 
	ppb_shifted = (((curr_int * si514_29bit_val + curr_frac) * ref_dec - common_div) * si514_Ppb_divisor)/ ((ref_int * si514_29bit_val + ref_frac) * curr_dec) */

	/* If the sum is negative, the mod is different
	 Normally the sum = result_int * common_div + result_frac
	 if the sum and result_int are both negative, this will substract from result_frac */
	if(sum > 0)
	{
		result_int = div128_with_rest(sum, common_div, &result_frac);
		result_frac = result_frac  * fraction_scale;
		result_frac = div128(result_frac, common_div);
	}
	else
	{
		sum = sum * (-1);
		result_int = (-1) * div128_with_rest(sum, common_div, &result_frac);
		result_frac = result_frac * fraction_scale;
		result_frac = (-1) * div128(result_frac, common_div);
	}

	*ppb_int_part = (__s32) result_int;
	*ppb_fract_part = (__s32) result_frac;
}

#endif /* _PPB_SHIFT_H */