
/****************************************************************************
 National Instruments Corporation.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 ****************************************************************************/

/*
 * Si514 Clock kernel driver
 */



/* si514 data sheet
https://www.silabs.co/Support%20Documents/TechnicalDocs/Si514.pdf
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/i2c.h> /* i2c device communications */
#include <linux/kobject.h>
#include <linux/sysfs.h>

#include <linux/slab.h>
#include <linux/device.h>
#include <asm/page.h>

#include <linux/types.h>
#include <generated/uapi/linux/version.h>
#include <linux/moduleparam.h>

#include "si514_regmap.h"
#include "ppbshift.h"

/****************************************************************************
 * Functions Declaration *
 ****************************************************************************/
static int si514_probe(struct i2c_client *client,
	const struct i2c_device_id *id);
static int si514_remove(struct i2c_client *client);

/* ------------------- Device Specific functions --------------------- */
static int si514_initial_configuration(struct i2c_client *client);
static int si514_update_multidiv_soft_copy(struct i2c_client *client);
static int si514_set_small_tune(struct i2c_client *client,
	const struct tSi514_muldiv *new_hw_settings);
static int si514_read_status(const struct i2c_client *client,
	struct tSi514_muldiv *settings);
static int si514_set_status(const struct i2c_client *client,
	const struct tSi514_muldiv *settings);
static int si514_trigger_status(const struct i2c_client *client,
	const struct tSi514_muldiv *settings);
static int si514_match_hw_status(const struct tSi514_muldiv *settings01,
	const struct tSi514_muldiv *settings02);
static int si514_match_hw_trigger_status(const struct tSi514_muldiv *settings01,
	const struct tSi514_muldiv *settings02);

/**********************
 * Device ID table *
 **********************/
static const struct i2c_device_id si514_idtable[] = {
	{"si514", 0},
	{}
};

MODULE_DEVICE_TABLE(i2c, si514_idtable);

/***********************
 * Device Driver *
 ***********************/
static struct i2c_driver si514_driver = {
	.driver = {
		.name = "si514",
		.owner = THIS_MODULE,
	},
	.id_table = si514_idtable,
	.probe = si514_probe,
	.remove = si514_remove,
};

/************************* Device Attributes *****************************/
static ssize_t si514_show_ppb_offset(struct device *dev,
	struct device_attribute *dev_attr,
	char *buf)
{
	s32 ppb_int = 0;
	s32 ppb_fract = 0;
	struct i2c_client *client;
	struct si514_data *dev_data;

	client = to_i2c_client(dev);
	dev_data = i2c_get_clientdata(client);

	mutex_lock(&dev_data->lock);
	compute_current_ppb_shift(dev_data, &ppb_int, &ppb_fract);
	mutex_unlock(&dev_data->lock);

	/* The return value is float, represented as two long numbers
	 * seperated (formatted) by dot.
	 * the first one is The integer part and the second one is
	 * the fraction as a part of 1000.
	*/

	return scnprintf(buf, PAGE_SIZE, "%ld.%ld\n", (long) ppb_int,
			(long) ppb_fract);
}

static ssize_t si514_store_ppb_offset(struct device *dev,
	struct device_attribute *dev_attr,
	const char *buf, size_t count)
{
	s32 shift_ppb_int;
	s32 shift_ppb_fract;
	struct tSi514_muldiv new_freq_settings;
	int ret;
	struct i2c_client *client;
	struct si514_data *dev_data;

	client = to_i2c_client(dev);
	dev_data = i2c_get_clientdata(client);

	/* The expected received value is float, represented as two long
	 * numbers seperated (formatted) by dot.
	 * the first one is The integer part and the second one is
	 * the fraction as a part of 1000.
	*/

	ret = sscanf(buf, "%ld.%ld", (long *) &shift_ppb_int,
			(long *) &shift_ppb_fract);

	if (ret != 2) {
		dev_warn(dev, "\n %s : Failed to read input parameter",
			 __func__);
		return -EIO;
	}

	mutex_lock(&dev_data->lock);
	ret = si514_compute_ppb_shift(client,
		&new_freq_settings, shift_ppb_int, shift_ppb_fract);
	if (ret < 0)
		goto err;
		/* User mode code will detect that this call failed */

	/* Shift the frequency */
	ret = si514_set_small_tune(client, &new_freq_settings);
	if (ret < 0)
		goto err;
		/* User mode code will detect that this call failed */

	ret = count;

err:
	mutex_unlock(&dev_data->lock);
	return ret;
}

static DEVICE_ATTR(ppb_offset, S_IWUSR | S_IWGRP | S_IRUGO,
	si514_show_ppb_offset, si514_store_ppb_offset);

/*----------------------------------------------------------------------*/
static ssize_t si514_show_fraction_scale(struct device *dev,
	struct device_attribute *dev_attr,
	char *buf)
{
return sprintf(buf, "%d\n", fraction_scale);
}
static DEVICE_ATTR(fraction_scale, S_IRUGO, si514_show_fraction_scale, NULL);
/*-----------------------------------------------------------------------*/

static struct attribute *si514_attributes[] = {
	&dev_attr_ppb_offset.attr, /*RW*/
	&dev_attr_fraction_scale.attr, /*R*/
	NULL
};

static struct attribute_group si514_attr_grp = {
	.attrs = si514_attributes,
};
/*///////////////////////////////////////////////////////////////////////////*/
module_param(max_allowed_ppb, int, 0644);
MODULE_PARM_DESC(max_allowed_ppb, "Maximum allowed ppb tuning value");

module_i2c_driver(si514_driver);

/*///////////////////////////////////////////////////////////////////////////*/

/*** Probe ***/
static int si514_probe(struct i2c_client *client,
	const struct i2c_device_id *id)
{
	int result;
	struct si514_data *si514_state_data;

	/* Initialize the driver data object */
	si514_state_data = NULL;
	si514_state_data = devm_kzalloc(&client->dev,
		sizeof(*si514_state_data), GFP_KERNEL);
	if (si514_state_data == NULL) {
		dev_err(&client->dev,
		"%s : failed to create driver data object\n", __func__);
		return -ENOMEM;
	}

	si514_state_data->i2c_client = client;
	i2c_set_clientdata(client, si514_state_data);

	/* Initialize Mutex */
	mutex_init(&si514_state_data->lock);

	/* Device Initial Configuration */
	result = si514_initial_configuration(client);
	if (result < 0) {
		/* Usually this failure due to HW communication */
		dev_err(&client->dev, "\n Failed initial configurations! ");
		goto error_mem;
	}

	/* Register device attributes to the sysfs */
	result = sysfs_create_group(&client->dev.kobj, &si514_attr_grp);
	if (result)
		goto error_mem;

	return 0;

error_mem:
	devm_kfree(&client->dev, si514_state_data);

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/
static int si514_remove(struct i2c_client *client)
{
	struct si514_data *dev_data;

	sysfs_remove_group(&client->dev.kobj, &si514_attr_grp);
	dev_data = i2c_get_clientdata(client);
	devm_kfree(&client->dev, dev_data);
	return 0;
}

/*///////////////////////////////////////////////////////////////////////////*/

static int si514_initial_configuration(struct i2c_client *client)
{
	int result;
	struct si514_data *dev_data_ptr;

	dev_data_ptr = i2c_get_clientdata(client);

	result = si514_update_multidiv_soft_copy(client);
	if (result < 0) {
		dev_err(&client->dev, "\n %s :failed to read HW registers\n",
			__func__);
		return result;
	}

	/*
	 * Detect which reference values we should be using based on
	 * hs_div setting. Since we do not adjust it in this driver,
	 * it should remain what it powered up with. The two reference
	 * values (25 and 100Mhz) this driver expects to deal with luckily
	 * have different hs_div values
	 */
	if (dev_data_ptr->current_hw_settings.hs_div ==
		k100MhzReferenceHWSettings.hs_div) {
		dev_data_ptr->reference_hw_settings =
			&k100MhzReferenceHWSettings;
	} else if (dev_data_ptr->current_hw_settings.hs_div ==
		k25MhzReferenceHWSettings.hs_div) {
		dev_data_ptr->reference_hw_settings =
			&k25MhzReferenceHWSettings;
	} else {
		dev_err(&client->dev,
		"%s :unknown value for hs_div\n", __func__);
		return -EIO;
	}

	/* Shift the frequency to default */
	result = si514_set_small_tune(client,
		dev_data_ptr->reference_hw_settings);
	if (result < 0)
		return result; /* fail initialization */

	/*
	 * Read the HW to validate we are in good state before moving
	 * forward
	 */
	result = si514_update_multidiv_soft_copy(client);
	if (result < 0) {
		dev_err(&client->dev, "\n %s :failed to read HW registers\n",
			__func__);
		return result;
	}

	/* at this point we are expecting the state of the HW registers
	is the same as the reference HW settings. Validate that */
	if (dev_data_ptr->reference_hw_settings->m_frac !=
		dev_data_ptr->current_hw_settings.m_frac ||
		dev_data_ptr->reference_hw_settings->m_int !=
		dev_data_ptr->current_hw_settings.m_int) {
		dev_warn(&client->dev,
			"\n %s : Reference settings doesn't match the initial HW state\n",
			__func__);
		return -EIO;
	}

return 0;
}

/*///////////////////////////////////////////////////////////////////////////*/
static int si514_update_multidiv_soft_copy(struct i2c_client *client)
{
	struct si514_data *dev_data;
	int result;

	result = 0;
	dev_data = i2c_get_clientdata(client);

	result = si514_read_status(client, &dev_data->current_hw_settings);

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

/* Configure the clock settings on the hardware */
static int si514_set_small_tune(struct i2c_client *client,
	const struct tSi514_muldiv *new_hw_settings)
{
	/* The si514 controls many clocks on the system. We
	   need to make sure that we don't put it in bad state as much
	   as we can.
	   The logic in this function as the following:
	   - write the desired values to the HW, but with out triggering the
		frequency update.
	   - read back and make sure they match what we already wrote.
	   - If it match, then trigger the frequency change. If not,
		try to write one more time (just to avoid any short and
		temporary noise on the bus "if any").
	   - After trigging the change, read the HW status one more time
		to make sure it match all the values have been written.
	   - if it didn't match, then abort and dont update the current
		HW settings, the user code can detect that something went
		wrong as the requested frequency didn't honored.
	*/

	int result;
	struct tSi514_muldiv settings;
	struct si514_data *dev_data;

	/* update the HW and trigger the changes */
	result = si514_set_status(client, new_hw_settings);
	if (result < 0)
		return result;

	/* read back status and make sure the status as has been set */
	result = si514_read_status(client, &settings);
	if (result < 0)
		return result;

	if (si514_match_hw_status(new_hw_settings, &settings) < 0) {
		/* data write fail. Try one more time */
		result = si514_set_status(client, new_hw_settings);
		if (result < 0)
			return result;

		if (si514_match_hw_status(new_hw_settings, &settings) < 0) {
			dev_err(&client->dev, "\n Failed to tune new frequency. IO error ");
			return -EIO;
		}
	}

	/* Trigger the frequency update */
	result = si514_trigger_status(client, new_hw_settings);
	if (result < 0)
		return result;

	/* read back status and make sure the status as has been set */
	result = si514_read_status(client, &settings);
	if (result < 0)
		return result;

	if (si514_match_hw_trigger_status(new_hw_settings, &settings) < 0) {
		dev_err(&client->dev, "\n Failed to tune new frequency. IO error ");
		return -EIO;
	}

	/* Reaching this point, means that the frequency update completed
	 * successfuly. Updat ethe current HW settings */
	 dev_data = i2c_get_clientdata(client);
	 dev_data->current_hw_settings = *new_hw_settings;

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

static int si514_set_status(const struct i2c_client *client,
	const struct tSi514_muldiv *settings)
{
	int result;
	/* Write all fine-tuning registers except the trigger register.
	the trigger registyer is at SI514_REG_OFFSET_M_INT offset,
	which cause the device to updat the frequency */
	u8 reg_buff[4];

	result = 0;

	reg_buff[0] = settings->m_frac & 0xff;
	reg_buff[1] = (settings->m_frac >> 8) & 0xff;
	reg_buff[2] = (settings->m_frac >> 16) & 0xff;
	reg_buff[3] = ((settings->m_frac >> 24) |
			(settings->m_int << 5)) & 0xff;

	result = i2c_smbus_write_i2c_block_data(client,
		(u8) SI514_REG_OFFSET_M_FRAC1, (u8) ARRAY_SIZE(reg_buff),
		reg_buff);

	if (result < 0) {
		/* Don't update the current HW settings
		 * At this stage the user mod code will detect that there
		 * is something wrong and will take action depends on that.
		 * At the driver side, this is a HW issue that we cant do
		 * too much about.*/
		dev_err(&client->dev, "%s : failed to write registers\n",
			__func__);
	}

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

static int si514_trigger_status(const struct i2c_client *client,
	const struct tSi514_muldiv *settings)
{
	u8 m_int;
	int result;

	m_int = (settings->m_int >> 3) & 0xff;

	result = i2c_smbus_write_byte_data(client, (u8) SI514_REG_OFFSET_M_INT,
		(u8) m_int);

	if (result < 0) {
		dev_err(&client->dev, "%s : failed to write trigger registers\n",
			__func__);
		return result;
	}

	udelay(100); /* Small tuning settling time. 100 micro second */
	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

static int si514_read_status(const struct i2c_client *client,
	struct tSi514_muldiv *settings)
{
	int result;
	u8 reg_buff[7];

	result = i2c_smbus_read_i2c_block_data(client,
		(u8) SI514_REG_OFFSET_M_FRAC1, (u8) ARRAY_SIZE(reg_buff),
		reg_buff);

	if (result < 0) {
		dev_err(&client->dev, "%s : failed to read registers\n",
			__func__);
		return result;
	}

	settings->m_frac = reg_buff[0] | ((u32) reg_buff[1] << 8) |
		((u32) reg_buff[2] << 16) |
		((u32) (reg_buff[3] & 0x1F) << 24);
	settings->m_int = ((u16) (reg_buff[4] & 0x3f) << 3) |
		(reg_buff[3] >> 5);
	settings->ls_div_bits = (reg_buff[6] >> 4) & 0x07;
	settings->hs_div = ((u16) (reg_buff[6] & 0x03) << 8) | reg_buff[5];

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

int si514_match_hw_status(const struct tSi514_muldiv *settings01,
	const struct tSi514_muldiv *settings02)
{
	if (settings01->m_frac == settings02->m_frac)
		return 0;

	return -1;
}

/*///////////////////////////////////////////////////////////////////////////*/

int si514_match_hw_trigger_status(const struct tSi514_muldiv *settings01,
	const struct tSi514_muldiv *settings02)
{
	if (settings01->m_int == settings02->m_int)
		return 0;

	return -1;
}

/*///////////////////////////////////////////////////////////////////////////*/

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Driver for Si514 Oscillators, I2C Driver");
MODULE_AUTHOR("Bassam Alsanie <bassam.alsanie@ni.com>");
